/*
 * The following is used to read the existing content from CS:GO compiled assets
 *
 * Supported formats:
 *   vdf
 *   vpk
 *   mdl
 */

#ifndef CXR_VALVE_BIN_H
#define CXR_VALVE_BIN_H

#include <stdlib.h>
#include <string.h>

#include "cxr_types.h"
#include "cxr_math.h"
#include "cxr_io.h"
#include "cxr_mem.h"
#include "cxr_log.h"

typedef struct VPKHeader VPKHeader;
typedef struct VPKDirectoryEntry VPKDirectoryEntry;
typedef struct vdf_kv vdf_kv;
typedef struct vdf_node vdf_node;
typedef struct vdf_ctx vdf_ctx;
typedef struct fs_locator fs_locator;

/*
 * These are 'unofficial' representations of the original formats, more C
 * friendly 
 */
typedef struct valve_file_system valve_file_system;
typedef struct valve_model valve_model;
typedef struct valve_model_batch valve_model_batch;
typedef struct valve_material valve_material;

/* Api */

CXR_API i32 cxr_fs_set_gameinfo( const char *path ); /* Setup system */
CXR_API void cxr_fs_exit(void);                       /* Clean up */
CXR_API void *cxr_fs_get( const char *path, i32 stringbuffer ); /* Get a file */
CXR_API void cxr_fs_free( void *data );
CXR_API i32 cxr_fs_find( const char *path, fs_locator *locator );

CXR_API valve_model *valve_load_model( const char *relpath );
CXR_API void valve_free_model( valve_model *model );
CXR_API valve_material *valve_load_material( const char *path );
CXR_API void valve_free_material( valve_material *material );

/*
 * File system implementation
 */

#pragma pack(push, 1)

struct VPKHeader
{
	u32 Signature;
	u32 Version;
	u32 TreeSize;
	u32 FileDataSectionSize;
	u32 ArchiveMD5SectionSize;
	u32 OtherMD5SectionSize;
	u32 SignatureSectionSize;
};

struct VPKDirectoryEntry
{
	u32 CRC;
	u16 PreloadBytes; 
	u16 ArchiveIndex;
	u32 EntryOffset;
	u32 EntryLength;
	u16 Terminator;
};

#pragma pack(pop)

static struct valve_file_system
{
   char *gamedir,
        *exedir;

   /* Runtime */
   VPKHeader vpk;
   char *directory_tree;

   cxr_abuffer searchpaths;

   FILE *current_archive;
   u16   current_idx;

   int  initialized;
}
fs_global = { .initialized = 0 };

struct fs_locator
{
   VPKDirectoryEntry *vpk_entry;
   char path[ 1024 ];
};

/* 
 * VPK reading
 */

static VPKDirectoryEntry *vpk_find( const char *asset );

/*
 * VDF reading
 */

static vdf_node *vdf_open_file( const char *fn );
static void vdf_free_r( vdf_node *p );

/* Starting from *it, get next child with matching name from node. */
static vdf_node *vdf_next( vdf_node *node, const char *name, int *it );

/* Create new empty node attached to parent. name can be NULL */
static vdf_node *vdf_create_node( vdf_node *parent, const char *name );

/* KV access */
static const char *kv_get( vdf_node *node, const char *key, 
      const char *value_defalt );

/* Iterate each keyvalue starting from *it until key is matched */
static char *kv_next( vdf_node *node, const char *key, int *it );

static int kv_get_int( vdf_node *node, const char *key, const int fallback );
static float kv_get_float( vdf_node *node, const char *key, float fallback );

static void kv_int_array( vdf_node *node, const char *key, u32 count, 
      int *arr );
static void kv_float_array( vdf_node *node, const char *key, u32 count, 
      float *arr );
static void kv_double_array( vdf_node *node, const char *key, u32 count, 
      double *arr );

#define vdf_foreach( NODE, STR, AS ) \
int __vdf_it_##AS = 0; \
vdf_node * AS;\
while( (AS = vdf_next( NODE, STR, &__vdf_it_##AS )) )

#define kv_foreach( NODE, STR, AS ) \
int __kv_it_##AS = 0; \
const char * AS;\
while( (AS = kv_next( NODE, STR, &__kv_it_##AS )) )

static VPKDirectoryEntry *vpk_find( const char *asset )
{
   valve_file_system *fs = &fs_global;

   if( !fs->directory_tree )
      return NULL;

   if( !asset )
      return NULL;
	
	char wbuf[ 512 ];
	strcpy( wbuf, asset );
	   
   /* 
    * This currently fails if the filename doesn't have a file extension, or
    * if it is located at the root path. I'm not sure if this is defined
    * behaviour or not. TODO: allow it to work anyway
    */

	char *ext = cxr_findext( wbuf, '.' );
   
   if( !ext ) return NULL;
	*(ext-1) = 0x00;

	char *fn = cxr_findext( wbuf, '/' );

   if( !fn ) return NULL;
	*(fn-1) = 0x00;

	char *dir = wbuf;
	char *pCur = fs->directory_tree;
	
	while( 1 )
	{
		if( !*pCur ) break;

		int bExt = !strcmp( ext, pCur );		
		
		while( *( pCur ++ ) ) {};
		while( 1 )
		{
			if( !*pCur ) { pCur ++; break; }

			int bDir = !strcmp( dir, pCur );

			while( *( pCur ++ ) ) {};
			while( 1 )
			{
				if( !*pCur ) { pCur ++; break; }
			
				const char *vpk_fn = pCur;
			
				while( *( pCur ++ ) ) {};
				VPKDirectoryEntry *entry = (VPKDirectoryEntry *)pCur;
				
				if( !strcmp( vpk_fn, fn ) && bExt && bDir ) 
				{
					return entry;
				}
				
				pCur += entry->PreloadBytes + sizeof( VPKDirectoryEntry );
			}
			
			if( bDir && bExt ) return NULL;
		}
		
		if( bExt ) return NULL;
	}
	
	return NULL;
}

/*
 * VDF 
 */

struct vdf_kv
{
	char *key;
	char *value;
};

struct vdf_node
{
	char *name;
	
	vdf_node *parent;
	
   cxr_abuffer abnodes,
               abpairs;
	
	u32		user;
	u32		user1;
};

static vdf_node *vdf_next( vdf_node *node, const char *name, int *it )
{	
	if( !node ) 
		return NULL;

	for( int i = it? *it: 0; i < node->abnodes.count; i ++ )
	{
      vdf_node **ptr_child = cxr_ab_ptr( &node->abnodes, i ),
                *child = *ptr_child;

		if( !name || !strcmp( name, child->name ))
		{
			if( it ) *it = i+1;
			return child;
		}
	}
	
	return NULL;
}

static char *kv_next( vdf_node *node, const char *key, int *it )
{
	char *val;

	if( node )
	{
		while( *it < node->abpairs.count )
		{
         vdf_kv *kv = cxr_ab_ptr( &node->abpairs, *it );

			if( !strcmp( kv->key, key ) )
			{
				val = kv->value;
				*it = *it + 1;
				return val;
			}
			
			*it = *it + 1;
		}
	}
	
	return NULL;
}

static const char *kv_get( vdf_node *node, const char *key, 
      const char *value_defalt )
{
   int it = 0;
   char *val = kv_next( node, key, &it );

   if( val )
      return val;

   return value_defalt;
}

static void vdf_str_to_int( const char *src, void *dest )
{
	*((int *)dest) = atoi( src );
}

static void vdf_str_to_float( const char *src, void *dest )
{
	*((float *)dest) = atof( src );
}

static void vdf_str_to_double( const char *src, void *dest )
{
	*((double *)dest) = atof( src );
}

static void kv_parse_array( const char *source, u32 esize, u32 count, 
      void(*interp_func)(const char *src, void *dest), void *arr )
{
	if( !source )
		return;

	char value_buf[ 64 ];
	int token = 0;
	
	u32 i = 0;
	u32 k = 0;
	
	char const *c = source;
	
	while( *c )
	{
		if( *c==' ' || *c=='\t' || *c=='[' || *c==']' || *c=='(' || *c==')' )
		{
			if( token )
			{
				value_buf[ k ] = 0x00;
				token = 0;
				
				interp_func( value_buf, ((u8 *)arr) + i*esize );
				i ++;
				
				if( i >= count )
				{
					break;
				}
			}
		}
		else
		{
			if( !token )
			{
				token = 1;
				k = 0;
			}
			
			if( token )
			{
				if( k < sizeof( value_buf ) - 1 )
				{
					value_buf[ k ++ ] = *c;
				}
			}
		}
	
		c ++;
	}
	
	/* Add remaining case if we hit null */
	if( token && (i < count) )
	{
		value_buf[ k ] = 0x00;
		interp_func( value_buf, ((u8 *)arr) + i*esize );
	}
}

static void kv_int_array( vdf_node *node, const char *key, u32 count, int *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(int), count, 
         vdf_str_to_int, arr );
}

static void kv_float_array( vdf_node *node, const char *key, u32 count, 
      float *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(float), count, 
         vdf_str_to_float, arr );
}

static void kv_double_array( vdf_node *node, const char *key, u32 count, 
      double *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(double), count, 
         vdf_str_to_double, arr );
}

static int kv_get_int( vdf_node *node, const char *key, 
      const int default_value )
{
	const char *v = kv_get( node, key, NULL );
	return v? atoi(v): default_value;
}

static float kv_get_float( vdf_node *node, const char *key, 
      float default_value )
{
	const char *v = kv_get( node, key, NULL );
	return v? atof( v ): default_value;
}

static vdf_node *vdf_create_node( vdf_node *parent, const char *name )
{
	vdf_node *node = calloc( 1, sizeof( vdf_node ) );
	
   /* init buffers */
   cxr_ab_init( &node->abnodes, sizeof( vdf_node* ), 0 );
   cxr_ab_init( &node->abpairs, sizeof( vdf_kv ), 0 );

	if( name )
	{
      node->name = cxr_str_clone(name, 0);
	}
	
	if( parent )
	{
		node->parent = parent;
		
      vdf_node **child = cxr_ab_empty( &parent->abnodes );
		*child = node;
	}
	
	return node;
}

static void vdf_kv_append( vdf_node *p, const char *k, const char *v )
{
   vdf_kv *kv = cxr_ab_empty( &p->abpairs );

	u32 sv = strlen(v)+1;
	u32 sk = strlen(k)+1;
	
	kv->key = malloc( sv+sk );
	kv->value = kv->key+sk;
	
	memcpy( kv->key, k, sk );
	memcpy( kv->value, v, sv );
}

static void vdf_free_r( vdf_node *p )
{
	for( int i = 0; i < p->abpairs.count; i ++ )
	{
      vdf_kv *kv = cxr_ab_ptr( &p->abpairs, i );
		free( kv->key );  /* key and value are allocated in the same buffer */
	}

	for( int i = 0; i < p->abnodes.count; i ++ )
	{
      vdf_node **ptr_node = cxr_ab_ptr( &p->abnodes, i );
		vdf_free_r( *ptr_node );
	}
	
   cxr_ab_free( &p->abpairs );
   cxr_ab_free( &p->abnodes );

	free( p->name );
	free( p );
}

/* 
 * Parsing 
 */

struct vdf_ctx
{
	char name[1024];
	
	vdf_node *root;
	
	u32 lines;
	u32 errors;
	
	struct 
	{
		int wait;
		int expect_decl;

		char *tokens[2];
		int i;
		
		char *ptr_read;
		
		vdf_node *pnode;
	} 
	st;
};

static void vdf_newln( vdf_ctx *ctx )
{
	ctx->lines ++;

	ctx->st.tokens[0] = NULL;
	ctx->st.tokens[1] = NULL;
	ctx->st.i = 0;
}

static void vdf_endl( vdf_ctx *ctx )
{
	if( ctx->st.tokens[0] )
	{
		/* Keypair */
		if( ctx->st.tokens[1] )
		{
			vdf_kv_append( ctx->st.pnode, ctx->st.tokens[0], ctx->st.tokens[1] );
		}
		else
		{
         /* decl */
			strcpy( ctx->name, ctx->st.tokens[0] );
			ctx->st.expect_decl = 1;
		}
	}
	
	vdf_newln( ctx );
}

static int vdf_line_control( vdf_ctx *ctx )
{
	if( *ctx->st.ptr_read == '\r' )
	{
		*ctx->st.ptr_read = 0x00;
		return 1;
	}
	if( *ctx->st.ptr_read == '\n' )
	{
		*ctx->st.ptr_read = 0x00;
		vdf_endl( ctx );
		return 2;
	}
	
	return 0;
}

static void vdf_wait_endl( vdf_ctx *ctx )
{
	while( (*ctx->st.ptr_read) && (*ctx->st.ptr_read != '\n') )
	{
		if( vdf_line_control( ctx ) == 2 )
		{
			return;
		}
	
		ctx->st.ptr_read ++;
	}
}

static void vdf_parse_string( vdf_ctx *ctx )
{
	while( *ctx->st.ptr_read )
	{
		if( *ctx->st.ptr_read == '"' )
		{
			*ctx->st.ptr_read = 0x00;
			return;
		}
		
		if( vdf_line_control( ctx ) )
		{
         /* Unexpected end of line */
         cxr_log( "vdf: unexpected EOL\n" );
			return;
		}
	
		ctx->st.ptr_read ++;
	}
}

static int vdf_parse_structure( vdf_ctx *ctx )
{
	if( *ctx->st.ptr_read == '{' )
	{
		if( ctx->st.tokens[0] || !ctx->st.expect_decl )
		{
         /* Unexpected token '{' */
         cxr_log( "vdf: Unexpected token '{'\n" );
			ctx->errors ++;
		}
		
		ctx->st.expect_decl = 0;
		ctx->st.pnode = vdf_create_node( ctx->st.pnode, ctx->name );
		
		vdf_wait_endl( ctx );
		return 1;
	}
	
	if( *ctx->st.ptr_read == '}' )
	{
		if( !ctx->st.pnode->parent )
		{
         /* Unexpected token '}' */
         cxr_log( "vdf: Unexpected token '}'\n" );
			ctx->errors ++;
		}
		else
		{
			ctx->st.pnode = ctx->st.pnode->parent;
		}
		
		vdf_wait_endl( ctx );
		return 1;
	}
	
	return 0;
}

static void vdf_parse_begin_token( vdf_ctx *ctx, char *ptr )
{
	ctx->st.tokens[ ctx->st.i ] = ptr;

	if( ctx->st.expect_decl )
	{
      /* Unexpected token 'name' */
      cxr_log( "vdf: unexpected token 'name'\n" );
		ctx->errors ++;
	}
}

static void vdf_parse_feedbuffer( vdf_ctx *ctx, char *buf )
{
	ctx->st.ptr_read = buf;
	
	while( *ctx->st.ptr_read )
	{
		if( !vdf_line_control( ctx ) )
		{
			if( (*ctx->st.ptr_read == '/') && (ctx->st.ptr_read[1] == '/') )
			{
				*ctx->st.ptr_read = 0x00;
				ctx->st.ptr_read += 2;
				
				vdf_endl( ctx );
				vdf_wait_endl( ctx );
			}
			else
			{
				if( !vdf_parse_structure( ctx ) )
				{
					if( *ctx->st.ptr_read == ' ' || *ctx->st.ptr_read == '\t' )
					{
						*ctx->st.ptr_read = 0x00;

						if( ctx->st.tokens[ ctx->st.i ] )
						{
							ctx->st.i ++;
							
							if( ctx->st.i == 2 ) 
							{
								vdf_wait_endl( ctx );
							}
						}
					}

               /* New entry */
					else if( !ctx->st.tokens[ ctx->st.i ] )
					{
						if( *ctx->st.ptr_read == '"' )
						{
							*ctx->st.ptr_read = 0x00;
							ctx->st.ptr_read ++;							
							
							vdf_parse_begin_token( ctx, ctx->st.ptr_read );
							vdf_parse_string( ctx );
						}
						else
						{
							if( !( *ctx->st.ptr_read == '/' && 
                              *(ctx->st.ptr_read + 1) == *ctx->st.ptr_read ) )
							{	
								vdf_parse_begin_token( ctx, ctx->st.ptr_read );
							}
						}
					}
				}
			}
		}
		
		ctx->st.ptr_read ++;
	}
}

static void vdf_debug_indent( int level )
{
   for(int i=0; i<level; i++)
      cxr_log(" ");
}

static void vdf_debug_r( vdf_node *node, int level )
{
   vdf_debug_indent(level);
   cxr_log( "vdf_node(%p, name: '%s')\n", node, node->name );

   vdf_debug_indent(level);
   cxr_log( "{\n" );

   for( int i=0; i<node->abpairs.count; i++ )
   {
      vdf_kv *kv = cxr_ab_ptr( &node->abpairs, i );

      vdf_debug_indent(level+1);
      cxr_log( "vdf_kv(%p, k: '%s', v: '%s')\n",
         kv, kv->key, kv->value );
   }

   for( int i=0; i<node->abnodes.count; i++ )
   {
      vdf_node **child = cxr_ab_ptr( &node->abnodes, i );
      vdf_debug_r( *child, level+1 );
   }

   vdf_debug_indent(level);
   cxr_log( "}\n" );
}

/* This will wreck the original buffer, but must stay alive! */
static vdf_node *vdf_from_buffer( char *buffer )
{
   vdf_node *root = vdf_create_node( NULL, NULL );

	vdf_ctx ctx = {0};
	ctx.root = ctx.st.pnode = root;
	
	vdf_newln( &ctx );
	vdf_parse_feedbuffer( &ctx, buffer );
   
#if 0
   vdf_debug_r( root, 0 );
#endif

   return root;
}

static vdf_node *vdf_open_file( const char *fn )
{	
	char *text_src = cxr_textasset_read( fn );
	
	if( !text_src )
		return NULL;

   vdf_node *root = vdf_from_buffer( text_src );
   free( text_src );

   return root;
}

/*
 * File system
 */

CXR_API i32 cxr_fs_set_gameinfo( const char *path )
{
   valve_file_system *fs = &fs_global;

   if( fs->initialized )
      cxr_fs_exit();

	vdf_node *info = vdf_open_file( path );
	if( !info ) 
		return 0;
	
   fs->gamedir = cxr_str_clone( path, 0 );
	cxr_downlvl( fs->gamedir );
	
   fs->exedir = cxr_str_clone( fs->gamedir, 0 );
	cxr_downlvl( fs->exedir );

   cxr_log( "Setting up file system:\n"
            "gameinfo: %s\n"
            "gamedir: %s\n"
            "exedir: %s\n",

            path, fs->gamedir, fs->exedir );
	
   /* get search paths */
	vdf_node *search_paths = 
      vdf_next
      (
         vdf_next
         (
            vdf_next( info, "GameInfo", NULL ), 
            "FileSystem", 
            NULL 
         ), 
         "SearchPaths", 
         NULL 
      );
	
   cxr_ab_init( &fs->searchpaths, sizeof( char *), 0 );

	kv_foreach( search_paths, "Game", kv_game )
	{
      cxr_log( "Game %s\n", kv_game );

		if( kv_game[0] == '|' ) continue;
	
		char *buf;
		if( cxr_path_is_abs( kv_game ) )
		{
         buf = cxr_str_clone( kv_game, 1 );
			strcat( buf, "/" );
		}
		else
		{
         buf = cxr_str_clone( fs->exedir, strlen(kv_game)+1 );
			strcat( buf, kv_game );
			strcat( buf, "/" );
		}
		
      char **sp = cxr_ab_empty( &fs->searchpaths );
      *sp = buf;
	}
	
	vdf_free_r( info );
	
   /* Find pack diretory */
	char pack_path[512];
   fs->current_archive = NULL;
   fs->current_idx = 0x7fff;

	for( int i = 0; i < fs->searchpaths.count; i ++ )
	{
      char **sp = cxr_ab_ptr( &fs->searchpaths, i );

		strcpy( pack_path, *sp );
		strcat( pack_path, "pak01_dir.vpk" );

      fs->current_archive = fopen( pack_path, "rb" );
   
      /* Read vpk directory */
      if( fs->current_archive )
      {
         fread( &fs->vpk, sizeof(VPKHeader), 1, fs->current_archive );

         fs->directory_tree = malloc( fs->vpk.TreeSize );
         fread( fs->directory_tree, fs->vpk.TreeSize, 1, fs->current_archive );

         break;
      }
	}
	
	if( !fs->current_archive )
	{
		cxr_log( "Could not locate pak01_dir.vpk in %i searchpaths. "
               "Stock models will not load!\n", fs->searchpaths.count );
	}
   
   fs->initialized = 1;
   return 1;
}

CXR_API void cxr_fs_exit(void)
{
	valve_file_system *fs = &fs_global;

	for( int i = 0; i < fs->searchpaths.count; i ++ )
	{
      char **sp = cxr_ab_ptr( &fs->searchpaths, i );
		free( *sp );
	}
   
   cxr_ab_free( &fs->searchpaths );

	if( fs->directory_tree )
	{
      free( fs->directory_tree );
      fs->directory_tree = NULL;
	}
	
	if( fs->current_archive )
	{
		fclose( fs->current_archive );
		fs->current_archive = NULL;
	}

   free( fs->gamedir );
   free( fs->exedir );
   
   memset( fs, 0, sizeof( valve_file_system ) );
}

static char *cxr_vpk_read( VPKDirectoryEntry *entry, int stringbuffer )
{
	valve_file_system *fs = &fs_global;

   if( !fs->initialized )
      return NULL;

   char pak[1024];
   
   /* Check if we need to change file handle */
   if( entry->ArchiveIndex != fs->current_idx )
   {
      if( fs->current_archive )
         fclose( fs->current_archive );

      fs->current_archive = NULL;
      fs->current_idx = entry->ArchiveIndex;

      if( entry->ArchiveIndex == 0x7fff )
      {
         snprintf( pak, 1023, "%scsgo/pak01_dir.vpk", fs->exedir );
         cxr_log( "locate %s\n", pak );
      }
      else
      {
         snprintf( pak, 1023, "%scsgo/pak01_%03hu.vpk", 
               fs->exedir, entry->ArchiveIndex );
         cxr_log( "locate %s\n", pak );
      }

      fs->current_archive = fopen( pak, "rb" );

      if( !fs->current_archive )
         cxr_log( "Warning: could not locate %s\n", pak );
   }

   if( !fs->current_archive )
      return NULL;

   size_t offset = entry->EntryOffset,
          length = entry->EntryLength;

   /*
    * File is stored in the index, after the tree
    */
   if( entry->ArchiveIndex == 0x7fff )
   {
      offset += fs->vpk.TreeSize + sizeof(VPKHeader);
   }
   
   /* 
    * Entire file is stored in the preload bytes;
    * Backtrack offset from directory to get absolute offset 
    */
   if( length == 0 )
   {
      offset = (char *)entry - (char *)fs->directory_tree;
      offset += sizeof( VPKHeader );
      offset += sizeof( VPKDirectoryEntry );

      length = entry->PreloadBytes;
   }
   else
      length += entry->PreloadBytes;

   fseek( fs->current_archive, offset, SEEK_SET );
   
   size_t alloc_size = stringbuffer? length+1: length;
   char *filebuf = malloc( alloc_size ),
        *fullfile = filebuf;

   if( stringbuffer )
      fullfile[length] = 0x00;

   if( entry->PreloadBytes )
   {
      memcpy( filebuf, ((char *)entry) + sizeof(VPKDirectoryEntry),
            entry->PreloadBytes );

      filebuf += entry->PreloadBytes;
      length -= entry->PreloadBytes;

      /* TODO, this can sometimes come from the same index file.
       *       very small performance gain can be made */
   }

   if( fread( filebuf, 1, length, fs->current_archive ) == length )
   {
      return fullfile;
   }
   else
   {
      /* Invalid read */
      free( filebuf );
      return NULL;
   }
}

CXR_API i32 cxr_fs_find( const char *path, fs_locator *locator )
{
	valve_file_system *fs = &fs_global;

   if( !fs->initialized )
      return 0;
	
	VPKDirectoryEntry *entry;

	if( fs->directory_tree )
	{
		if( (entry = vpk_find( path )) )
		{
         locator->vpk_entry = entry;
         locator->path[0] = 0x00;
         return 1;
		}
	}

   locator->vpk_entry = NULL;

   /* Use physical search paths */
	for( int i = 0; i < fs->searchpaths.count; i ++ )
	{
      char **sp = cxr_ab_ptr( &fs->searchpaths, i );

      snprintf( locator->path, 1024, "%s%s", *sp, path );
      
      if( cxr_file_exists( locator->path ) )
         return 1;
	}
	
   /* File not found */
	return 0;
}

CXR_API void cxr_fs_free( void *data )
{
   free( data );
}

CXR_API void *cxr_fs_get( const char *path, i32 stringbuffer )
{
	valve_file_system *fs = &fs_global;

   if( !fs->initialized )
      return NULL;

   fs_locator locator;

   if( cxr_fs_find( path, &locator ) )
   {
      if( locator.vpk_entry )
      {
         return cxr_vpk_read( locator.vpk_entry, stringbuffer );
      }
      else
      {
         char *filebuf;

         if( stringbuffer )
            return cxr_textasset_read( locator.path );
         else
            return cxr_asset_read( locator.path );
      }
   }

   return NULL;
}

/*
 * VMDL interface
 *
 * This software is not affiliated with Valve Corporation
 *   We are not affiliated, associated, authorized, endorsed by, or in any way 
 *   officially connected with Valve Corporation, or any of its subsidiaries or 
 *   its affiliates. 
 *
 *   All trademarks are property of their respective owners
 */

#define MAX_NUM_LODS 8
#define MAX_NUM_BONES_PER_VERT 3

#pragma pack(push, 1)
typedef struct
{
	float	weight[MAX_NUM_BONES_PER_VERT];
	char 	bone[MAX_NUM_BONES_PER_VERT];
	char 	numbones;
} 
boneWeight_t;

typedef struct
{
	boneWeight_t boneweights;
	float 		pos[3];
	float 		norm[3];	
	float 		uv[2];
} 
mstudiovertex_t;

typedef struct
{
	int id;
	int version;
	int checksum;
	int numLods;
	int numLodVertexes[MAX_NUM_LODS];
	int numFixups;
	int fixupTableStart;
	int vertexDataStart;
	int tangentDataStart;
} 
vertexFileHeader_t;

#pragma pack(pop)

static mstudiovertex_t *GetVertexData( vertexFileHeader_t *t ) 
{	
	return (mstudiovertex_t *) ( (char *)t + t->vertexDataStart );
}

/*
 * VTX file format
 */

#pragma pack(push, 1)

typedef struct
{
	unsigned char boneWeightIndex[3];
	unsigned char numBones;

	unsigned short origMeshVertID;
	char boneID[3];
} 
VTXVertex_t;

enum StripGroupFlags
{
	STRIPGROUP_IS_FLEXED = 0x01,
	STRIPGROUP_IS_HWSKINNED = 0x02,
	STRIPGROUP_IS_DELTA_FLEXED = 0x04,
	STRIPGROUP_SUPPRESS_HW_MORPH = 0x08
};

enum StripHeaderFlags_t {
	STRIP_IS_TRILIST	= 0x01,
	STRIP_IS_TRISTRIP	= 0x02      /* Unused by studiomdl 2015? */
};

typedef struct 
{
	int numIndices;
	int indexOffset;

	int numVerts;
	int vertOffset;

	short numBones;

	unsigned char flags;

	int numBoneStateChanges;
	int boneStateChangeOffset;
} 
VTXStripHeader_t;

typedef struct
{
	int numVerts;
	int vertOffset;

	int numIndices;
	int indexOffset;

	int numStrips;
	int stripOffset;

	unsigned char flags;
} 
VTXStripGroupHeader_t;

static VTXVertex_t *pVertexVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (VTXVertex_t *)(((char *)t) + t->vertOffset) + i;
}

static unsigned short *pIndexVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (unsigned short *)(((char *)t) + t->indexOffset) + i;
}

static VTXStripHeader_t *pStripVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (VTXStripHeader_t *)(((char *)t) + t->stripOffset) + i; 
}

typedef struct
{
	int numStripGroups;
	int stripGroupHeaderOffset;

	unsigned char flags;
} 
VTXMeshHeader_t;
static VTXStripGroupHeader_t *pStripGroupVTX( VTXMeshHeader_t *t, int i ) 
{ 
	return (VTXStripGroupHeader_t *)(((char *)t) + t->stripGroupHeaderOffset) +i; 
}

typedef struct
{
	int numMeshes;
	int meshOffset;

	float switchPoint;
} 
VTXModelLODHeader_t;
static VTXMeshHeader_t *pMeshVTX( VTXModelLODHeader_t *t, int i ) 
{ 
	return (VTXMeshHeader_t *)(((char *)t) + t->meshOffset) + i; 
}

typedef struct
{
	int numLODs;
	int lodOffset;
} 
VTXModelHeader_t;
static VTXModelLODHeader_t *pLODVTX( VTXModelHeader_t *t, int i ) 
{ 
	return (VTXModelLODHeader_t *)(((char *)t) + t->lodOffset) + i; 
}

typedef struct
{
	int numModels;
	int modelOffset;
} 
VTXBodyPartHeader_t;
static VTXModelHeader_t *pModelVTX( VTXBodyPartHeader_t *t, int i ) 
{ 
	return (VTXModelHeader_t *)(((char *)t) + t->modelOffset) + i;
}

typedef struct
{
	int version; /* 7 */

	/* hardware params that affect how the model is to be optimized. */
	int vertCacheSize;
	unsigned short maxBonesPerStrip;
	unsigned short maxBonesPerTri;
	int maxBonesPerVert;

	int checkSum;
	int numLODs;
	int materialReplacementListOffset;
	int numBodyParts;
	int bodyPartOffset;
} 
VTXFileHeader_t;
static VTXBodyPartHeader_t *pBodyPartVTX( VTXFileHeader_t *t, int i ) 
{
	return (VTXBodyPartHeader_t *)(((char *)t) + t->bodyPartOffset) + i;
}

/*
                .VTX file structure
   =============================================

   FileHeader
     L	BodyParts::
        L	Models::
           L	LODS::
              L	Meshes::
                 L	StripGroups::
                    L	VerticesTable[StudioMDL.Vertex]
                    L	IndicesTable[UINT16]
                    |
                    L	Strips::
                       L	Vertices[UINT16]
                       L	Indices[UINT16]
*/

#pragma pack(pop)

/* 
 * mdl format
 */

#pragma pack(push, 1)

typedef struct
{
	void			*pVertexData;
	void			*pTangentData;
} 
mstudio_modelvertexdata_t;

typedef struct
{
	int					unused_modelvertexdata;
	int					numLODVertexes[MAX_NUM_LODS];
	
	mstudio_modelvertexdata_t *_the_death_ptr;
} 
mstudio_meshvertexdata_t;

typedef struct mstudiomodel_t mstudiomodel_t;
typedef struct
{
	int					material;
	int					modelindex; 
	int					numvertices;
	int					vertexoffset;
	int					numflexes;	
	int					flexindex;
	int					materialtype;
	int					materialparam;
	int					meshid;
	float					center[3];
	mstudio_meshvertexdata_t vertexdata;
	int					unused[6]; 
} 
mstudiomesh_t;

struct mstudiomodel_t
{
	char					name[64];
	int					type;
	float					boundingradius;

	int					nummeshes;	
	int					meshindex;

	int					numvertices;
	int					vertexindex;
	int					tangentsindex;

	int					numattachments;
	int					attachmentindex;

	int					numeyeballs;
	int					eyeballindex;

	mstudio_modelvertexdata_t vertexdata;

	int					unused[8];
};
static mstudiomesh_t *studiomodel_pMesh( mstudiomodel_t *t, int i ) 
{ 
	return (mstudiomesh_t *)(((char *)t) + t->meshindex) + i; 
}

typedef struct
{
	int					sznameindex;
	int					nummodels;
	int					base;
	int					modelindex;
} mstudiobodyparts_t;

static mstudiomodel_t *mstudiobodyparts_pModel( mstudiobodyparts_t *t, int i ) 
{ 
	return (mstudiomodel_t *)(((char *)t) + t->modelindex) + i;
}

typedef struct
{
	int	id;
	int	version;
	int	checksum;
	char	name[64];
	int	length;
	float	eyeposition[3];
	float	illumposition[3];
	float	hull_min[3];
	float	hull_max[3];
	float	view_bbmin[3];
	float	view_bbmax[3];
	int	flags;
	int	numbones;
	int	boneindex;
	int	numbonecontrollers;
	int	bonecontrollerindex;
	int	numhitboxsets;
	int	hitboxsetindex;
	int	numlocalanim;		
	int	localanimindex;
	int	numlocalseq;
	int	localseqindex;
	int 	activitylistversion;
	int 	eventsindexed;
	int	numtextures;
	int	textureindex;
	int	numcdtextures;
	int	cdtextureindex;
	int	numskinref;
	int	numskinfamilies;
	int	skinindex;
	int	numbodyparts;
	int	bodypartindex;
	int	numlocalattachments;
	int	localattachmentindex;
	int	numlocalnodes;
	int	localnodeindex;
	int	localnodenameindex;
	int	numflexdesc;
	int	flexdescindex;
	int	numflexcontrollers;
	int	flexcontrollerindex;
	int	numflexrules;
	int	flexruleindex;
	int	numikchains;
	int	ikchainindex;
	int	nummouths;
	int	mouthindex;
	int	numlocalposeparameters;
	int	localposeparamindex;
	int	surfacepropindex;
	int	keyvalueindex;
	int	keyvaluesize;
	int	numlocalikautoplaylocks;
	int	localikautoplaylockindex;
	float	mass;
	int	contents;
	int	numincludemodels;
	int	includemodelindex;
	int	szanimblocknameindex;
	int	numanimblocks;
	int	animblockindex;
	int	bonetablebynameindex;
	char	constdirectionallightdot;
	char	rootLOD;
	char	numAllowedRootLODs;
	char	unused[1];
	int	unused4;
	int	numflexcontrollerui;
	int	flexcontrolleruiindex;
	float	flVertAnimFixedPointScale;
	int	unused3[1];
	int	studiohdr2index;
	int	unused2[1];
} 
studiohdr_t;

static char *studiohdr_pCdtexture( studiohdr_t *t, int i )
{ 
   return (((char *)t) + *((int *)(((u8 *)t) + t->cdtextureindex) + i)); 
}

static mstudiobodyparts_t *studiohdr_pBodypart( studiohdr_t *t, int i ) 
{ 
	return (mstudiobodyparts_t *)(((char *)t) + t->bodypartindex) + i; 
}

typedef struct
{
   int						sznameindex;
   int						flags;
   int						used;
   int                  unused1;
	int						garbage[12]; /* contains some studiomdl pointers */
} 
mstudiotexture_t;

static char *mstudiotexture_pszName( mstudiotexture_t *t )
{ 
   return ((char *)t) + t->sznameindex;
}

static mstudiotexture_t *studiohdr_pTexture( studiohdr_t *t, int i )
{ 
   return (mstudiotexture_t *)(((u8 *)t) + t->textureindex) + i; 
} 

#pragma pack(pop)

static void vtx_resource_counts( VTXFileHeader_t *pVtxHdr, studiohdr_t *pMdl,
      u32 *indices, u32 *meshes )
{
   *indices = 0;
   *meshes = 0;

	for( int bodyID = 0; bodyID < pVtxHdr->numBodyParts; ++bodyID )
	{
		/* Body parts */
		VTXBodyPartHeader_t* pVtxBodyPart = pBodyPartVTX( pVtxHdr, bodyID );
		mstudiobodyparts_t *pBodyPart = studiohdr_pBodypart( pMdl, bodyID );
		
		for( int modelID = 0; modelID < pBodyPart->nummodels; ++modelID )
		{	
         /* models */
			VTXModelHeader_t* pVtxModel = pModelVTX( pVtxBodyPart, modelID );
			mstudiomodel_t *pStudioModel = 
            mstudiobodyparts_pModel( pBodyPart, modelID );

			int nLod = 0;
			VTXModelLODHeader_t *pVtxLOD = pLODVTX( pVtxModel, nLod );

			for( int nMesh = 0; nMesh < pStudioModel->nummeshes; ++nMesh )
			{
            /* meshes */
				VTXMeshHeader_t* pVtxMesh = pMeshVTX( pVtxLOD, nMesh );
				mstudiomesh_t* pMesh = studiomodel_pMesh( pStudioModel, nMesh );

            (*meshes)++;

				for ( int nGroup = 0; nGroup < pVtxMesh->numStripGroups; ++nGroup )
				{
               /* groups */
					VTXStripGroupHeader_t* pStripGroup = 
                  pStripGroupVTX( pVtxMesh, nGroup );

					for ( int nStrip = 0; nStrip < pStripGroup->numStrips; nStrip++ )
					{
                  /* strips */
						VTXStripHeader_t *pStrip = pStripVTX( pStripGroup, nStrip );

						if ( pStrip->flags & STRIP_IS_TRILIST )
						{
                     *indices += pStrip->numIndices;
						}
					}
				}
			}
		}
	}
}

/* 
 * The following section is the wrappers for the underlying types
 */

struct valve_material
{
   char *basetexture,
        *bumpmap;
};

struct valve_model
{
   float *vertex_data;  /* pos xyz, norm xyz, uv xy */

   u32 *indices,
        indices_count,
        vertex_count,
        part_count,
        material_count;

   char **materials;

   struct valve_model_batch
   {
      u32 material,
          ibstart,
          ibcount;
   }
   *parts;

   /* Internal valve data */
   studiohdr_t *studiohdr;
   VTXFileHeader_t *vtxhdr;
   vertexFileHeader_t *vvdhdr;
};

CXR_API valve_model *valve_load_model( const char *relpath )
{
   char path[1024];
   strcpy( path, relpath );

   char *ext = cxr_stripext( path );

   if( !ext )
      return NULL;

   /* Load data files */
   valve_model *model = malloc( sizeof( valve_model ) );
   model->studiohdr = NULL;
   model->vtxhdr = NULL;
   model->vvdhdr = NULL;

   strcpy( ext, ".dx90.vtx" );
   model->vtxhdr = cxr_fs_get( path, 0 );

   strcpy( ext, ".vvd" );
   model->vvdhdr = cxr_fs_get( path, 0 );

   strcpy( ext, ".mdl" );
   model->studiohdr = cxr_fs_get( path, 0 );

   if( !model->vvdhdr || !model->studiohdr || !model->vtxhdr )
   {
      cxr_log( "Error, failed to load: (%s)\n", relpath );

      free( model->studiohdr );
      free( model->vvdhdr );
      free( model->studiohdr );
      free( model );

      return NULL;
   }

   /* allocate resources */
   vtx_resource_counts( model->vtxhdr, model->studiohdr, 
         &model->indices_count, &model->part_count );
   model->vertex_count = model->vvdhdr->numLodVertexes[0];
   model->material_count = model->studiohdr->numtextures;

   model->materials = malloc( model->material_count * sizeof(char *) );
   model->parts = malloc( sizeof( valve_model_batch ) * model->part_count );
   model->indices = malloc( sizeof( u32 ) * model->indices_count );
   model->vertex_data = malloc( sizeof( float ) * 8 * model->vertex_count );
   
   /* Find materials */
   for( int i=0; i<model->studiohdr->numtextures; i++ )
   {
      char material_path[ 1024 ];
      fs_locator locator;

      mstudiotexture_t *tex = studiohdr_pTexture( model->studiohdr, i );
      const char *name = mstudiotexture_pszName( tex );

      model->materials[i] = NULL;
      
      cxr_log( "search: %s\n", name );

      for( int j=0; j<model->studiohdr->numcdtextures; j++ )
      {
         char *cdpath = studiohdr_pCdtexture( model->studiohdr, j );
         snprintf( material_path, 1023, "materials/%s%s.vmt", cdpath, name );
         cxr_unixpath( material_path );
         cxr_lowercase( material_path );

         if( cxr_fs_find( material_path, &locator ))
         {
            model->materials[i] = cxr_str_clone( material_path, 0 );
            break;
         }
      }

   }

   u32 i_index = 0,
       i_mesh = 0;

   /* Extract meshes */
	for( int bodyID = 0; bodyID < model->studiohdr->numbodyparts; ++bodyID )
	{
		/* Body parts */
		VTXBodyPartHeader_t* pVtxBodyPart = pBodyPartVTX( model->vtxhdr, bodyID );
		mstudiobodyparts_t *pBodyPart = 
         studiohdr_pBodypart( model->studiohdr, bodyID );
		
		for( int modelID = 0; modelID < pBodyPart->nummodels; ++modelID )
		{	
         /* models */
			VTXModelHeader_t* pVtxModel = pModelVTX( pVtxBodyPart, modelID );
			mstudiomodel_t *pStudioModel = 
            mstudiobodyparts_pModel( pBodyPart, modelID );

			int nLod = 0;
			VTXModelLODHeader_t *pVtxLOD = pLODVTX( pVtxModel, nLod );

			for( int nMesh = 0; nMesh < pStudioModel->nummeshes; ++nMesh )
			{
            /* meshes, each of these creates a new draw CMD */
				VTXMeshHeader_t* pVtxMesh = pMeshVTX( pVtxLOD, nMesh );
				mstudiomesh_t* pMesh = studiomodel_pMesh( pStudioModel, nMesh );

            valve_model_batch *curBatch = &model->parts[ i_mesh ++ ];
            curBatch->material = pMesh->material;
            curBatch->ibstart = i_index;
            curBatch->ibcount = 0;

				for( int nGroup = 0; nGroup < pVtxMesh->numStripGroups; ++nGroup )
				{
               /* groups */
					VTXStripGroupHeader_t* pStripGroup = 
                  pStripGroupVTX( pVtxMesh, nGroup );

					for( int nStrip = 0; nStrip < pStripGroup->numStrips; nStrip++ )
					{
                  /* strips */
						VTXStripHeader_t *pStrip = pStripVTX( pStripGroup, nStrip );

						if( pStrip->flags & STRIP_IS_TRILIST )
						{
                     /* indices */
							for( int i = 0; i < pStrip->numIndices; i ++ )
							{
								u16 i1 = *pIndexVTX( pStripGroup, 
                              pStrip->indexOffset + i );								

								model->indices[ i_index ++ ] = 
                           pVertexVTX( pStripGroup, i1 )->origMeshVertID + 
                           pMesh->vertexoffset;

                        curBatch->ibcount ++;
							}
						}
                  else
                  {
                     /* This is unused? */
                  }
					}
				}
			}
		}
	}
	
	mstudiovertex_t *vertexData = GetVertexData( model->vvdhdr );

	for( int i = 0; i < model->vertex_count; i ++ )
	{
		mstudiovertex_t *vert = &vertexData[i];
      
      float *dest = &model->vertex_data[ i*8 ];

      dest[0] = vert->pos[0];
      dest[1] = vert->pos[1];
      dest[2] = vert->pos[2];

      dest[3] = vert->norm[0];
      dest[4] = vert->norm[1];
      dest[5] = vert->norm[2];

      dest[6] = vert->uv[0];
      dest[7] = vert->uv[1];
	}
	
   return model;
}

CXR_API void valve_free_model( valve_model *model )
{
   for( int i=0; i<model->material_count; i++ )
      free( model->materials[i] );

   free( model->materials );
   free( model->parts );
   free( model->indices );
   free( model->vertex_data );

   free( model->studiohdr );
   free( model->vtxhdr );
   free( model->vvdhdr );
   free( model );
}

static char *valve_texture_path( const char *path )
{
   if( !path )
      return NULL;

   char *buf = 
      malloc( strlen( path ) + strlen(".vtf") + strlen("materials/") +1 );
   strcpy( buf, "materials/" );
   strcat( buf, path );
   strcat( buf, ".vtf" );

   cxr_unixpath( buf );
   cxr_lowercase( buf );

   return buf;
}

static void cxr_str_debug( const char *str )
{
   cxr_log( "\n=== string debug ( %p ) === \n'", str );
   const char *y = str;

   while(*y)
   {
      if( *y == '\r' )
         cxr_log( "<CR>" );
      else if( *y == '\n' )
         cxr_log( "<LF>\n" );
      else
         cxr_log( "%c", *y);
      y ++;
   }
   cxr_log( "<EOF>'\n" );
}

CXR_API valve_material *valve_load_material( const char *path )
{
   char *vmt = cxr_fs_get( path, 1 );

   if( vmt )
   {
#if 0
      cxr_str_debug( vmt );
#endif
      valve_material *material = malloc( sizeof(valve_material) );
      vdf_node *vmt_root = vdf_from_buffer( vmt );

      if( vmt_root->abnodes.count == 0 )
      {
         cxr_log( "Error: vmt has no nodes\n" );
         free( vmt );
         free( material );
         vdf_free_r( vmt_root );
         return NULL;
      }
      
      vdf_node **body = cxr_ab_ptr( &vmt_root->abnodes, 0 );

      /* Path semantics here are inconsistent
       * I believe they should all just be converted to lowercase, though */

      for( int i=0; i<(*body)->abpairs.count; i++ )
      {
         vdf_kv *kv = cxr_ab_ptr( &(*body)->abpairs, i );
         cxr_lowercase( kv->key );
      }
      
      const char *basetexture = kv_get( *body, "$basetexture", NULL ),
                 *bumpmap = kv_get( *body, "$bumpmap", NULL );
      
      /* TODO: other shader parameters */
      material->basetexture = valve_texture_path( basetexture );
      material->bumpmap = valve_texture_path( bumpmap );
      
      vdf_free_r( vmt_root );
      free(vmt);

      return material;
   }

   return NULL;
}

CXR_API void valve_free_material( valve_material *material )
{
   free( material->basetexture );
   free( material->bumpmap );
   free( material );
}

#endif /* CXR_VALVE_BIN_H */
