// nbvtf.h - v1.03 - Writer for Valve Texture Format - public domain
// Written by Harry 'hgn' Godden
//
// Credits:
//   Rich Geldreich - bc7enc (BC1/BC3 High Quality texture compression)
//   Fabian "ryg" Giesen, stb - stb_dxt (BC1/BC3 realtime compressors)
//   Sean Barrett - stb_image.h, stb_image_write.h (Image I/O)
//
// Note:
//   This library assumes normal maps are input in OpenGL(correct) format, and will be converted into DirectX(incorrect) at
//   compile time automatically. Do not submit DirectX normals into this software.
//
// Since this project uses stb_image, use '#define STB_IMAGE_IMPLEMENTATION' before including
// Additionally, to use high quality DXT, '#define USE_LIBRGBCX'
//
// USAGE:
//   Call these functions:
//     int nbvtf_convert( const char *src, int w, int h, int mipmap, EImageFormat_t format, uint32_t usr_flags, const char *dest );
//     int nbvtf_write( uint8_t *src, int w, int h, int mipmap, EImageFormat_t format, uint32_t usr_flags, const char *dest );
//
// Param definitions:
//   src - RGBA byte data of image
//   w - width of image
//   h - height of image
//   mipmap - enable mipmap generation (box filter), 1/0
//   format - Choose from: k_EImageFormat_DXT1, compressedk_EImageFormat_DXT5, k_EImageFormat_BGR888, k_EImageFormat_ABGR8888
//   usr_flags - You can append any flags but only really need TEXTUREFLAGS_NORMAL if texture is normal map
//   dest - file path to write vtf to
//   qual - Image quality 0-18 (rgbcx, stb always highqual)
//  
// Convert specific:
//   src - file path of source image to convert
//   w, h - MAXIMUM image dimentions of final product. Set as 0 to be automatic
//
// version history:
//   v1.03  - Added quality switch, moved init (major api revision)
//   v1.02  - Improved box filtering, small bug fixes
//   v1.01  - switch to OpenGL normal format for input
//   v1.00  - (hgn) first release
//
// LICENSE
//   See end of file for license information.

#ifdef __cplusplus
extern "C" {
#endif

#define NBVTF_MAX_MIPLEVELS 9
#define nbvtf__min(a,b) (((a)<(b))?(a):(b))
#define nbvtf__max(a,b) (((a)>(b))?(a):(b))

#ifdef NBVTF_AS_SO
 #include <stdio.h>
 #include <string.h>
 #include <stdint.h>
 #include <math.h>
 
 #define NBVTF_SHOW_STDERR
 #define STB_IMAGE_IMPLEMENTATION
 #define STB_IMAGE_WRITE_IMPLEMENTATION
#endif

#define STBI_NO_THREAD_LOCALS
#include "stb/stb_image.h"
#include "stb/stb_image_write.h"

#ifdef USE_LIBRGBCX
 // #define RGBCX_NO_ALGORITHM
 #include "librgbcx.h"
 #define HAS_DXT_COMPRESSOR
#endif

#if USE_STB_DXT
 #define STB_DXT_IMPLEMENTATION
 #include "stb/stb_dxt.h"
 #define HAS_DXT_COMPRESSOR
#endif

#ifndef HAS_DXT_COMPRESSOR
 #warning No DXT compressor specified! S3 Output will be invalid.
#endif

#ifdef NBVTF_SHOW_STDERR
 #define NBVTF_ERR(...)
#else
 #define NBVTF_ERR(...) fprintf( stderr, __VA_ARGS__ )
#endif

#pragma pack(push, 1)

typedef enum EImageFormat
{
   /* Format                           Export   Import */
	k_EImageFormat_NONE = -1,
	k_EImageFormat_RGBA8888 = 0,	   // -        yes
	k_EImageFormat_ABGR8888,         // yes      yes
	k_EImageFormat_RGB888,				// -        yes
	k_EImageFormat_BGR888,           // yes      yes
	k_EImageFormat_RGB565,           // -        planned
	k_EImageFormat_I8,               // -        planned
	k_EImageFormat_IA88,             // -        planned
	k_EImageFormat_P8,               // -        -
	k_EImageFormat_A8,               // -        -
	k_EImageFormat_RGB888_BLUESCREEN,// -        -
	k_EImageFormat_BGR888_BLUESCREEN,// -        -
	k_EImageFormat_ARGB8888,         // -        yes
	k_EImageFormat_BGRA8888,         // -        yes
	k_EImageFormat_DXT1,					// yes      yes
	k_EImageFormat_DXT3,             // -        -
	k_EImageFormat_DXT5,					// yes      yes
	k_EImageFormat_BGRX8888,         // -        -
	k_EImageFormat_BGR565,           // -        planned
	k_EImageFormat_BGRX5551,         // -        -
	k_EImageFormat_BGRA4444,         // -        -
	k_EImageFormat_DXT1_ONEBITALPHA, // -        planned
	k_EImageFormat_BGRA5551,         // -        -
	k_EImageFormat_UV88,             // -        -
	k_EImageFormat_UVWQ8888,         // -        -
	k_EImageFormat_RGBA16161616F,    // -        -
	k_EImageFormat_RGBA16161616,     // -        -
	k_EImageFormat_UVLX8888          // -        -
} EImageFormat_t;

const char *vtf_format_strings[] = 
{
	"RGBA8888",
	"ABGR8888",
	"RGB888",
	"BGR888",
	"RGB565",
	"I8",
	"IA88",
	"P8",
	"A8",
	"RGB888_BLUESCREEN",
	"BGR888_BLUESCREEN",
	"ARGB8888",
	"BGRA8888",
#ifdef USE_LIBRGBCX
	"DXT1 (rgbcx.h)",
#else
	"DXT1 (stb_dxt.h)",
#endif
	"DXT3",
#ifdef USE_LIBRGBCX
	"DXT5 (rgbcx.h)",
#else
	"DXT5 (stb_dxt.h)",
#endif
	"BGRX8888",
	"BGR565",
	"BGRX5551",
	"BGRA4444",
	"DXT1_ONEBITALPHA",
	"BGRA5551",
	"UV88",
	"UVWQ8888",
	"RGBA16161616F",
	"RGBA16161616",
	"UVLX8888"
};

enum flag
{
	// Flags from the *.txt config file
	TEXTUREFLAGS_POINTSAMPLE = 0x00000001,
	TEXTUREFLAGS_TRILINEAR = 0x00000002,
	TEXTUREFLAGS_CLAMPS = 0x00000004,
	TEXTUREFLAGS_CLAMPT = 0x00000008,
	TEXTUREFLAGS_ANISOTROPIC = 0x00000010,
	TEXTUREFLAGS_HINT_DXT5 = 0x00000020,
	TEXTUREFLAGS_PWL_CORRECTED = 0x00000040,
	TEXTUREFLAGS_NORMAL = 0x00000080,
	TEXTUREFLAGS_NOMIP = 0x00000100,
	TEXTUREFLAGS_NOLOD = 0x00000200,
	TEXTUREFLAGS_ALL_MIPS = 0x00000400,
	TEXTUREFLAGS_PROCEDURAL = 0x00000800,

	// These are automatically generated by vtex from the texture data.
	TEXTUREFLAGS_ONEBITALPHA = 0x00001000,
	TEXTUREFLAGS_EIGHTBITALPHA = 0x00002000,

	// Newer flags from the *.txt config file
	TEXTUREFLAGS_ENVMAP = 0x00004000,
	TEXTUREFLAGS_RENDERTARGET = 0x00008000,
	TEXTUREFLAGS_DEPTHRENDERTARGET = 0x00010000,
	TEXTUREFLAGS_NODEBUGOVERRIDE = 0x00020000,
	TEXTUREFLAGS_SINGLECOPY = 0x00040000,
	TEXTUREFLAGS_PRE_SRGB = 0x00080000,

	TEXTUREFLAGS_UNUSED_00100000 = 0x00100000,
	TEXTUREFLAGS_UNUSED_00200000 = 0x00200000,
	TEXTUREFLAGS_UNUSED_00400000 = 0x00400000,

	TEXTUREFLAGS_NODEPTHBUFFER = 0x00800000,

	TEXTUREFLAGS_UNUSED_01000000 = 0x01000000,

	TEXTUREFLAGS_CLAMPU = 0x02000000,
	TEXTUREFLAGS_VERTEXTEXTURE = 0x04000000,
	TEXTUREFLAGS_SSBUMP = 0x08000000,

	TEXTUREFLAGS_UNUSED_10000000 = 0x10000000,

	TEXTUREFLAGS_BORDER = 0x20000000,

	TEXTUREFLAGS_UNUSED_40000000 = 0x40000000,
	TEXTUREFLAGS_UNUSED_80000000 = 0x80000000,
};

typedef struct vtfheader
{
	union
	{
	char				signature[4];			// File signature ("VTF\0"). (or as little-endian integer, 0x00465456)
	uint32_t			usig;
	};
	
	unsigned int	version[2];				// version[0].version[1] (currently 7.2).
	unsigned int	headerSize;				// Size of the header struct  (16 byte aligned; currently 80 bytes) + size of the resources dictionary (7.3+).
	unsigned short	width;					// Width of the largest mipmap in pixels. Must be a power of 2.
	unsigned short	height;					// Height of the largest mipmap in pixels. Must be a power of 2.
	unsigned int	flags;					// VTF flags.
	unsigned short	frames;					// Number of frames, if animated (1 for no animation).
	unsigned short	firstFrame;				// First frame in animation (0 based).
	unsigned char	padding0[4];			// reflectivity padding (16 byte alignment).
	float				reflectivity[3];		// reflectivity vector.
	unsigned char	padding1[4];			// reflectivity padding (8 byte packing).
	float				bumpmapScale;			// Bumpmap scale.
	unsigned int	highResImageFormat;	// High resolution image format.
	unsigned char	mipmapCount;			// Number of mipmaps.
	unsigned int	lowResImageFormat;	// Low resolution image format (always DXT1).
	unsigned char	lowResImageWidth;		// Low resolution image width.
	unsigned char	lowResImageHeight;	// Low resolution image height.

	// 7.2+
	unsigned short	depth;					// Depth of the largest mipmap in pixels.
													// Must be a power of 2. Can be 0 or 1 for a 2D texture (v7.2 only).

	// 7.3+
	unsigned char	padding2[3];			// depth padding (4 byte alignment).
	unsigned int	numResources;			// Number of resources this vtf has
	
	unsigned char  padding3[8];        // Necessary on certain compilers
} vtfheader_t;

#pragma pack(pop)

#pragma pack(push, 1)
struct DDS_PIXELFORMAT
{
	uint32_t dwSize;
	uint32_t dwFlags;
	uint32_t dwFourCC;
	uint32_t dwRGBBitCount;
	uint32_t dwRBitMask;
	uint32_t dwGBitMask;
	uint32_t dwBBitMask;
	uint32_t dwABitMask;
};

struct DDS_HEADER {
	uint32_t           dwSize;
	uint32_t           dwFlags;
	uint32_t           dwHeight;
	uint32_t           dwWidth;
	uint32_t           dwPitchOrLinearSize;
	uint32_t           dwDepth;
	uint32_t           dwMipMapCount;
	uint32_t           dwReserved1[11];
	struct DDS_PIXELFORMAT	ddspf;
	uint32_t           dwCaps;
	uint32_t           dwCaps2;
	uint32_t           dwCaps3;
	uint32_t           dwCaps4;
	uint32_t           dwReserved2;
};

#pragma pack(pop)

uint32_t swap_endian(uint32_t val)
{
	return (val << 24) | ((val << 8) & 0x00ff0000) |
		   ((val >> 8) & 0x0000ff00) | (val >> 24);
}

#define DDSD_CAPS 0x1
#define DDSD_HEIGHT 0x2
#define DDSD_WIDTH 0x4
#define DDSD_PITCH 0x8
#define DDSD_PIXELFORMAT 0x1000
#define DDSD_MIPMAPCOUNT 0x20000
#define DDSD_LINEARSIZE 0x80000
#define DDSD_DEPTH 0x800000

#define DDPF_ALPHAPIXELS 0x1
#define DDPF_ALPHA 0x2
#define DDPF_FOURCC 0x4
#define DDPF_RGB 0x40
#define DDPF_YUV 0x200
#define DDPF_LUMINANCE 0x20000

#define DDSCAPS_COMPLEX 0x8
#define DDSCAPS_MIPMAP 0x400000
#define DDSCAPS_TEXTURE 0x1000

#define BLOCK_SIZE_DXT1 8
#define BLOCK_SIZE_DXT5 16

#define BBP_RGB888 24
#define BBP_RGBA8888 32

#define DDS_HEADER_SIZE 124
#define DDS_HEADER_PFSIZE 32
#define DDS_MAGICNUM 0x20534444;

#define DDS_FLIP_VERTICALLY_ON_WRITE

typedef struct mipimg
{
	uint32_t w;
	uint32_t h;
	uint32_t src_offset;
} mipimg_t;

int nbvtf_power2( uint32_t x )
{
    return (x != 0) && ((x & (x - 1)) == 0);
}

int nbvtf_power2x( uint32_t y, uint32_t x )
{
	return nbvtf_power2( y ) && nbvtf_power2( x );
}

int nbvtf_lower( int *x, int *y )
{
	if( *x == 1 && *y == 1 )
	{
		return 0;
	}

	*x = nbvtf__max( 1, (*x)/2 );
	*y = nbvtf__max( 1, (*y)/2 );
	
	return 1;
}

int nbvtf_lowres_index( int w, int h )
{
	int x, y;
	int i = 0;
	
	x = w;
	y = h;
	
	while(1)
	{
		if( (x <= 16) && ( y <= 16 ) )
		{
			return i;
		}
		
		i ++;
		
		nbvtf_lower( &x, &y );
	}
}

// Simple box filter downscale
void nbvtf_downscale( uint8_t *src, int w, int h, int dw, int dh, uint8_t *dest )
{
   int bx = w/dw;
   int by = h/dh;
   int div = bx*by;

	for( int y = 0; y < dh; y ++ )
   for( int x = 0; x < dw; x ++ )
   {
      // Average block colours
      uint32_t tr = 0, tg = 0, tb = 0, ta = 0;

      for( int yy = 0; yy < by; yy ++ )
      for( int xx = 0; xx < bx; xx ++ )
      {
         uint8_t *psrc = &src[ (x*bx+xx + (y*by+yy)*w)*4 ];
         tr+=psrc[0];
         tg+=psrc[1];
         tb+=psrc[2];
         ta+=psrc[3];
      }
      
      uint8_t *pdst = &dest[ (y*dw + x)*4 ];
      pdst[0] = tr / div;
      pdst[1] = tg / div;
      pdst[2] = tb / div;
      pdst[3] = ta / div;
	}
}

uint8_t *nbvtf_create_mipmaps( uint8_t *src, int w, int h, mipimg_t *offsets, int *num )
{
	int memory = 0;
	int x, y, i;
	uint32_t offset;
	
	x = w;
	y = h;
	while( nbvtf_lower( &x, &y ) )
		memory += x*y*4;
	
	uint8_t *mipmem = (uint8_t *)malloc( memory );
	
	if( mipmem )
	{
		x = w;
		y = h;
		i = 0;
		offset = 0;
		
		uint8_t *dest = mipmem;
		
		while(1)
		{
			if( !nbvtf_lower( &x, &y ) )
				break;

			nbvtf_downscale( src, w, h, x, y, dest );
			
			offsets[ i ].src_offset = offset;
			offsets[ i ].w = x;
			offsets[ i ].h = y;
			i ++;
			
			offset += x*y*4;
			dest = mipmem + offset;
		}
		
		*num = i;
		return mipmem;
	}
	else
	{
		NBVTF_ERR( "nbvtf_write:err out of memory allocating mipmap buffer!\n" );
		return NULL;
	}
}

void nbvtf_reflectivity( uint8_t *src, int w, int h, float *dest )
{
	uint32_t totals[3] = {0,0,0};
	
	for( int i = 0; i < w*h; i ++ )
	{
		totals[0] += src[i*4+0];
		totals[1] += src[i*4+1];
		totals[2] += src[i*4+2];
	}
	
	dest[0] = (float)( totals[0] / (w*h) ) / 255.0f;
	dest[1] = (float)( totals[1] / (w*h) ) / 255.0f;
	dest[2] = (float)( totals[2] / (w*h) ) / 255.0f;
}

#ifdef NBVTF_ALLOW_EXPORT
void nbvtf_debug_view_mips( uint8_t *src, int w, int h )
{
	int x, y, i;
	char fnbuf[512];
	
	x = w;
	y = h;
	i = 1;

	uint8_t *dest = src;

	while( nbvtf_lower( &x, &y ) )
	{
		sprintf( fnbuf, "mip_%d.png", i ++ );
	
		stbi_write_png( fnbuf, x,y, 4, dest, x*4 );
		dest += x*y*4;
	}
}
#endif

void nbvtf_dxt_pad( uint8_t *src, int bx, int by, int w, int h, uint8_t *dest )
{
	int px = bx*4;
	int py = by*4;
	
	uint32_t *stream = (uint32_t *)src;
	uint32_t *stream_out = (uint32_t *)dest;

	for( int y = 0; y < 4; y ++ )
	{
		for( int x = 0; x < 4; x ++ )
		{
			stream_out[ y*4+x ] = stream[ nbvtf__min( py+y, h-1 )*w + nbvtf__min( px+x, w-1 ) ];
		}
	}
}

uint32_t nbvtf_dxt_sizeimg( int w, int h, int alpha )
{
	uint32_t blocks_x, blocks_y;
	int block_size = alpha? 16: 8;

	blocks_x = ((uint32_t)w) >> 2;
	blocks_y = ((uint32_t)h) >> 2;
	
	int padx = w % 4 != 0? 1: 0;
	int pady = h % 4 != 0? 1: 0;
	
	return (blocks_x+padx)*(blocks_y+pady)*block_size;
}

uint32_t nbvtf_sizeimg( int w, int h, EImageFormat_t format )
{
	if( format == k_EImageFormat_DXT5 || format == k_EImageFormat_DXT1 )
	{
		return nbvtf_dxt_sizeimg( w, h, format == k_EImageFormat_DXT1? 0: 1 );
	}
	
	if( format == k_EImageFormat_BGR888 )
		return w*h*3;
		
	if( format == k_EImageFormat_ABGR8888 )
		return w*h*4;
		
	return 0;
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
void nbvtf_init(void)
{
#ifdef USE_LIBRGBCX
   rgbcx__init();
#endif
}

void nbvtf_dxt_block( uint8_t *dest, uint8_t *src, int alpha, int qual )
{
#ifdef USE_LIBRGBCX
	if( alpha )
	{
		rgbcx__encode_bc3( qual, dest, src );
	}
	else
	{
		rgbcx__encode_bc1( qual, dest, src, 0, 0 );
	}
#endif

#if USE_STB_DXT
	stb_compress_dxt_block( dest, src, alpha, STB_DXT_HIGHQUAL );
#endif

#ifndef HAS_DXT_COMPRESSOR

   for( int i=0; i<alpha? 16: 8; i++ )
      dest[i] = i%1? 0xff: 0x00;
#endif
}

void nbvtf_dxt_block_unpack( uint8_t *src, uint8_t *dest, int alpha )
{
#ifdef USE_LIBRGBCX
	if( alpha )
	{
		rgbcx__unpack_bc3( src, dest );
	}
	else
	{
		rgbcx__unpack_bc1( src, dest );
	}
#endif

#if USE_STB_DXT
   stb_decompress_dxt_block( src, dest, alpha );
#endif

#ifndef HAS_DXT_COMPRESSOR
   for( int i=0; i<alpha? 128: 64; i++ )
      dest[i] = i%1? 0xff: 0x00;
#endif
}

void nbvtf_compress_dxt( uint8_t *src, int w, int h, int alpha, int qual,
      uint8_t *dest )
{
	uint32_t blocks_x, blocks_y;

	blocks_x = ((uint32_t)w) >> 2;
	blocks_y = ((uint32_t)h) >> 2;
	
	int padx = w % 4 != 0? 1: 0;
	int pady = h % 4 != 0? 1: 0;
	
	int block_size = alpha? 16: 8;
	
	uint8_t *dest_block = dest;
	
	uint8_t working_block[ 4*4*4 ];
	
	// Compress rows
	for( int y = 0; y < blocks_y; y ++ )
	{
		for( int x = 0; x < blocks_x; x ++ )
		{
			uint8_t *src_begin = src + (y*w*4 + x*4)*4;
			for( int i = 0; i < 4; i ++ )
			{
				memcpy( working_block + i*4*4, src_begin + w*4*i, 4*4 );
			}
			
			nbvtf_dxt_block( dest_block, working_block, alpha, qual );
			dest_block += block_size;
		}
		
		if( padx )
		{
			nbvtf_dxt_pad( src, blocks_x, y, w, h, working_block );
			nbvtf_dxt_block( dest_block, working_block, alpha, qual );
			dest_block += block_size;
		}
	}
	
	// Compress remainder row
	if( pady )
	{
		for( int x = 0; x < blocks_x; x ++ )
		{
			nbvtf_dxt_pad( src, x, blocks_y, w, h, working_block );
			nbvtf_dxt_block( dest_block, working_block, alpha, qual );
			dest_block += block_size;
		}
	}
	
	// Compress last little corner
	if( padx && pady )
	{
		nbvtf_dxt_pad( src, blocks_x, blocks_y, w, h, working_block );
		nbvtf_dxt_block( dest_block, working_block, alpha, qual );
	}
}

/* Decompress block to rgba, padding currently doesn't get read */
void nbvtf_decompress_dxt(uint8_t *src, int w, int h, int alpha, uint8_t *dest )
{
	uint32_t blocks_x, blocks_y;

	blocks_x = ((uint32_t)w) >> 2;
	blocks_y = ((uint32_t)h) >> 2;
	
	int padx = w % 4 != 0? 1: 0;
	int pady = h % 4 != 0? 1: 0;
	
	int block_size = alpha? 16: 8;
	
	uint8_t *dxt_block = src;
	uint8_t working_block[ 4*4*4 ];
	
	// Compress rows
	for( int y = 0; y < blocks_y; y ++ )
	{
		for( int x = 0; x < blocks_x; x ++ )
		{
         nbvtf_dxt_block_unpack( dxt_block, working_block, alpha );

			uint8_t *dest_begin = dest + (y*w*4 + x*4)*4;
			for( int i = 0; i < 4; i ++ )
				memcpy( dest_begin + w*4*i, working_block + i*4*4, 4*4 );
			
			dxt_block += block_size;
		}
		
		if( padx )
			dxt_block += block_size;
	}
	
	if( pady )
		for( int x = 0; x < blocks_x; x ++ )
			dxt_block += block_size;
	
	if( padx && pady )
	{
	}
}
static void nbvtf_swizzle_to( uint8_t *src, int n, int nc, uint8_t *dest )
{
	for( int i = 0; i < n; i ++ )
	{
		for( int j = 0; j < nc; j ++ )
		{
			dest[ i*nc+nc-j-1 ] = src[ i*4+j ];
		}
	}
}

static void nbvtf_write_img_data( uint8_t *src, int w, int h, 
      EImageFormat_t format, int qual, uint8_t *wb, FILE *file )
{
	switch( format )
	{
		case k_EImageFormat_DXT1:
			nbvtf_compress_dxt( src, w, h, 0, qual, wb );
			fwrite( wb, nbvtf_dxt_sizeimg( w, h, 0 ), 1, file );
		break;
		case k_EImageFormat_DXT5:
			nbvtf_compress_dxt( src, w, h, 1, qual, wb );
			fwrite( wb, nbvtf_dxt_sizeimg( w, h, 1 ), 1, file );
		break;
		case k_EImageFormat_ABGR8888:
			nbvtf_swizzle_to( src, w*h, 4, wb );
			fwrite( wb, w*h*4, 1, file );
		break;
		case k_EImageFormat_BGR888:
			nbvtf_swizzle_to( src, w*h, 3, wb );
			fwrite( wb, w*h*3, 1, file );
		break;
		
		default:
		break;
	}
}

static size_t nbvtf_img_size( int w, int h, EImageFormat_t format )
{
   int block_count = nbvtf__max(1, ((w + 3) / 4)) * 
                     nbvtf__max(1, ((h + 3) / 4));

   switch( format )
   {
      case k_EImageFormat_RGBA8888:
      case k_EImageFormat_ABGR8888:
      case k_EImageFormat_ARGB8888:
      case k_EImageFormat_BGRA8888:
         return 4*w*h;

      case k_EImageFormat_RGB888:
      case k_EImageFormat_BGR888:
         return 3*w*h;

      case k_EImageFormat_RGB565:
      case k_EImageFormat_IA88:
         return 2*w*h;

      case k_EImageFormat_I8:
         return w*h;
      
      case k_EImageFormat_DXT1:
         return block_count * BLOCK_SIZE_DXT1;

      case k_EImageFormat_DXT5:
         return block_count * BLOCK_SIZE_DXT5;

      default:
         break;
   }
}

static void nbvtf_read_img_data( uint8_t *src, int w, int h, 
      EImageFormat_t format, uint8_t *dst )
{
   switch( format )
   {
      case k_EImageFormat_RGBA8888:
         for( int i=0; i<w*h*4; i++ )
            dst[i] = src[i];
         break;

      case k_EImageFormat_ABGR8888:
         for( int i=0; i<w*h; i++ )
         {
            dst[i*4+0] = src[i*4+3];
            dst[i*4+1] = src[i*4+2];
            dst[i*4+2] = src[i*4+1];
            dst[i*4+3] = src[i*4+0];
         }
         break;

      case k_EImageFormat_RGB888:
         for( int i=0; i<w*h; i++ )
         {
            dst[i*4+0] = src[i*3+0];
            dst[i*4+1] = src[i*3+1];
            dst[i*4+2] = src[i*3+2];
            dst[i*4+3] = 0xFF;
         }
         break;

      case k_EImageFormat_BGR888:
         for( int i=0; i<w*h; i++ )
         {
            dst[i*4+0] = src[i*3+2];
            dst[i*4+1] = src[i*3+1];
            dst[i*4+2] = src[i*3+0];
            dst[i*4+3] = 0xFF;
         }
         break;

      case k_EImageFormat_RGB565:
      case k_EImageFormat_I8:
      case k_EImageFormat_IA88:
         /* Planned */
         break;
      
      case k_EImageFormat_ARGB8888:
         for( int i=0; i<w*h; i++ )
         {
            dst[i*4+0] = src[i*4+1];
            dst[i*4+1] = src[i*4+2];
            dst[i*4+2] = src[i*4+3];
            dst[i*4+3] = src[i*4+0];
         }
         break;

      case k_EImageFormat_BGRA8888:
         for( int i=0; i<w*h; i++ )
         {
            dst[i*4+0] = src[i*4+2];
            dst[i*4+1] = src[i*4+1];
            dst[i*4+2] = src[i*4+0];
            dst[i*4+3] = src[i*4+3];
         }
         break;
      
      case k_EImageFormat_DXT1:
         nbvtf_decompress_dxt( src, w, h, 0, dst );
         break;

      case k_EImageFormat_DXT5:
         nbvtf_decompress_dxt( src, w, h, 1, dst );
         break;

      default:
         break;
   }
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
int nbvtf_write_dds_dxt1( uint8_t *reference, int w, int h, int qual, const char *dest )
{
	if( !nbvtf_power2x(w,h) )
	{
		NBVTF_ERR( "nbvtf_write:err image dimentions were not power of two (%d %d)\n", w, h );
		return 0;
	}

	struct DDS_HEADER header = {0};
	header.dwSize = DDS_HEADER_SIZE;
	header.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	header.dwHeight = h;
	header.dwWidth = w;
   header.dwPitchOrLinearSize = nbvtf__max(1, ((w + 3) / 4)) * BLOCK_SIZE_DXT1;
	header.ddspf.dwSize = DDS_HEADER_PFSIZE;
   header.ddspf.dwFlags |= DDPF_FOURCC;
   header.ddspf.dwFourCC = ((uint32_t)'D'<<0) |
                           ((uint32_t)'X'<<8) |
                           ((uint32_t)'T'<<16) | 
                           ((uint32_t)'1'<<24);

   header.dwFlags |= DDSD_LINEARSIZE;
	header.dwMipMapCount = 0;
	header.dwCaps = DDSCAPS_TEXTURE;

	// Magic number
	uint32_t magic = DDS_MAGICNUM;
   
   FILE *file = fopen( dest, "wb" );
   fwrite( &magic, sizeof(uint32_t), 1, file );
   fwrite( &header, DDS_HEADER_SIZE, 1, file );

	uint32_t size_highres = nbvtf_sizeimg( w, h, k_EImageFormat_DXT1 );
	uint8_t *working_buffer = malloc( size_highres );

   nbvtf_compress_dxt( reference, w, h, 0, qual, working_buffer );
   fwrite( working_buffer, size_highres, 1, file );

   free( working_buffer );
   fclose( file );
   return 1;
}

static void nbvtf_correct_normal( uint8_t *src, uint8_t *dst, int w, int h )
{
   for( int i=0; i < w*h; i++ )
   {
      dst[i*4+0] = src[i*4+0];
      dst[i*4+1] = 0xFF-src[i*4+1];
      dst[i*4+2] = src[i*4+2];
      dst[i*4+3] = src[i*4+3];
   }
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
int nbvtf_write( uint8_t *reference, int w, int h, int mipmap, 
      EImageFormat_t format, int qual, uint32_t usr_flags, const char *dest )
{
	if( !nbvtf_power2x(w,h) )
	{
		NBVTF_ERR( "nbvtf_write:err image dimentions were not power of two (%d %d)\n", w, h );
		return 0;
	}

	mipimg_t mip_offsets[ 16 ];
	int num_mips;
	
   uint8_t *src;

   // Convert to directx normal
   if( usr_flags & TEXTUREFLAGS_NORMAL )
   {
      src = malloc( w*h*4 );
      nbvtf_correct_normal( reference, src, w, h );
   }
   else
      src = reference;
	
	uint8_t *mip_data = nbvtf_create_mipmaps( src, w, h, mip_offsets, &num_mips );
	
	if( !mip_data )
	{
		NBVTF_ERR( "nbvtf_write:err mipmap data failed to generate" );

      if( usr_flags & TEXTUREFLAGS_NORMAL )
         free( src );

		return 0;
	}
	
	vtfheader_t header = {0};
	
	header.usig = 0x00465456;
	header.headerSize = sizeof( vtfheader_t );
	header.version[0] = 7;
	header.version[1] = 2;
	
	header.width = w;
	header.height = h;
	header.flags = usr_flags;
	
	// Append format flags
	if( !mipmap )
	{
		header.flags |= TEXTUREFLAGS_NOLOD;
		header.flags |= TEXTUREFLAGS_NOMIP;
	}
	
	if( format == k_EImageFormat_DXT5 || format == k_EImageFormat_ABGR8888 )
	{
		header.flags |= TEXTUREFLAGS_EIGHTBITALPHA;
	}
	
	header.frames = 1;
	header.firstFrame = 0;
	nbvtf_reflectivity( mip_data + mip_offsets[ num_mips-1 ].src_offset, 1,1, header.reflectivity );
	header.bumpmapScale = 1.0f;
	
	header.highResImageFormat = format;
	header.mipmapCount = mipmap? 
      nbvtf__min(num_mips,NBVTF_MAX_MIPLEVELS)+1: 1;
	
	header.lowResImageFormat = k_EImageFormat_DXT1;

	header.depth = 1;
	header.numResources = 0;
	
	int lr_index = nbvtf_lowres_index( w, h );
	
	uint8_t *lr_src;
	
	if( lr_index )
	{
		mipimg_t *mip = mip_offsets + (lr_index-1);
		lr_src = mip_data + mip->src_offset;
		
		header.lowResImageWidth = mip->w;
		header.lowResImageHeight = mip->h;
	}
	else
	{
		lr_src = src;
		
		header.lowResImageWidth = w;
		header.lowResImageHeight = h;
	}
	
	uint32_t size_highres = nbvtf_sizeimg( w, h, format );
	uint32_t size_lowres = 
      nbvtf_dxt_sizeimg( header.lowResImageWidth, header.lowResImageHeight, 0 );
	
	uint8_t *working_buffer = 
      (uint8_t *)malloc( nbvtf__max( size_highres, size_lowres ) );
	
	if( !working_buffer )
	{
		NBVTF_ERR( "nbvtf_write:err out of memory allocating working buffer\n" );
		free( mip_data );

      if( usr_flags & TEXTUREFLAGS_NORMAL )
         free( src );

		return 0;
	}
	
	FILE *file = fopen( dest, "wb" );
	
	if( !file )
	{
		NBVTF_ERR( "nbvtf_write:err could not open file stream for writing\n" );
		
		free( working_buffer );
		free( mip_data );
		
      if( usr_flags & TEXTUREFLAGS_NORMAL )
         free( src );

		return 0;
	}
	
	// Write header
	fwrite( &header, sizeof( vtfheader_t ), 1, file );
	
	// Write low res
	nbvtf_write_img_data( 
		lr_src, header.lowResImageWidth, header.lowResImageHeight, 
      k_EImageFormat_DXT1, qual, working_buffer, file 
	);
	
	// Write texture data
	if( mipmap )
	{
      // !! Experimental !!
      int start = nbvtf__max( 0, num_mips-NBVTF_MAX_MIPLEVELS );
      
		for( int i = start; i < num_mips; i ++ )
		{
			mipimg_t *mip = mip_offsets + (num_mips - i -1);
			nbvtf_write_img_data( mip_data + mip->src_offset, mip->w, mip->h, 
               format, qual, working_buffer, file );
		}
	}
	
	// Write high resolution
	nbvtf_write_img_data( src, w, h, format, qual, working_buffer, file );
	
	fclose( file );

	free( working_buffer );
	free( mip_data );

   if( usr_flags & TEXTUREFLAGS_NORMAL )
      free( src );
	
	return 1;
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
uint8_t *nbvtf_read( vtfheader_t *header, int32_t *w, int32_t *h, int normal )
{
   *w = header->width;
   *h = header->height;

   uint8_t *rgba = malloc( header->width * header->height * 4 );
   
   if( !rgba )
      return NULL;
   
	size_t memory = 0;
   int x = header->width,
       y = header->height;

   int mip_iter = 0;
	
	while( nbvtf_lower( &x, &y ) )
   {
      if( mip_iter == header->mipmapCount )
         break;

      mip_iter ++;
		memory += nbvtf_img_size( x, y, header->highResImageFormat );
   }

	if( header->lowResImageWidth == 0 || header->lowResImageHeight == 0 ||
         header->lowResImageFormat == 0xffffffff )
   {
      /* no thumbnail? */
   }
   else
      memory += nbvtf_img_size( header->lowResImageWidth,
                                header->lowResImageHeight,
                                header->lowResImageFormat );

   /* Decode high res image into rgba */
   nbvtf_read_img_data( ((uint8_t *)header) + header->headerSize + memory,
                        header->width, header->height,
                        header->highResImageFormat,
                        rgba );

   if( normal )
   {
      nbvtf_correct_normal( rgba, rgba, header->width, header->height );
      stbi_write_png( "/tmp/cxr_hello_n.png", header->width, header->height,
            4, rgba, header->width*4 );
   }
   else
      stbi_write_png( "/tmp/cxr_hello.png", header->width, header->height,
            4, rgba, header->width*4 );


   return rgba;
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
uint8_t *nbvtf_raw_data( vtfheader_t *header, int32_t *w, int32_t *h, int32_t *format )
{
   *w = header->width;
   *h = header->height;
   *format = header->highResImageFormat;

   return ((uint8_t *)header) + header->headerSize;
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
void nbvtf_free( uint8_t *data )
{
   free(data);
}

#ifdef NBVTF_AS_SO
__attribute__((visibility("default")))
#endif
int nbvtf_convert( const char *src, int w, int h, int mipmap, 
      EImageFormat_t format, int qual, uint32_t usr_flags, const char *dest )
{
	if( (w && h) && !nbvtf_power2x(w,h) )
	{
		NBVTF_ERR( "nbvtf_convert:err requested dimentions were not power of two (%d %d)\n", w, h );
		return 0;
	}
	
	int x,y,n;
	uint8_t *data = stbi_load( src, &x, &y, &n, 4 );
	
	if( data )
	{
		if( !nbvtf_power2x(x,y) )
		{
			NBVTF_ERR( "nbvtf_convert:err loaded image dimentions were not power two (%d %d)\n", x, y );
			stbi_image_free( data );
			return 0;
		}
		
		// Image size needs to be made smaller
		if( (w && h) && ( x > w || y > h ) )
      {
         nbvtf_downscale( data, x, y, w, h, data );
         x = w;
         y = h;
      }
		
		int status = nbvtf_write( data, x,y, mipmap, format, qual, usr_flags, dest );

		stbi_image_free( data );
		
		return status;
	}
	else
	{
		return 0;
	}
}

#ifdef __cplusplus
}
#endif

/*
LICENSE - Applies to nbvtf.h, pynbvtf.py, librgbcx.cc, librgbcx.h, vtf_cmd.c
------------------------------------------------------------------------------
Public Domain (www.unlicense.org)
This is free and unencumbered software released into the public domain.
Anyone is free to copy, modify, publish, use, compile, sell, or distribute this
software, either in source code form or as a compiled binary, for any purpose,
commercial or non-commercial, and by any means.
In jurisdictions that recognize copyright laws, the author or authors of this
software dedicate any and all copyright interest in the software to the public
domain. We make this dedication for the benefit of the public at large and to
the detriment of our heirs and successors. We intend this dedication to be an
overt act of relinquishment in perpetuity of all present and future rights to
this software under copyright law.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
------------------------------------------------------------------------------
*/
